#ifndef ETSI_GS_QKD_004_RINGBUFFER_H
#define ETSI_GS_QKD_004_RINGBUFFER_H

#include <time.h>

/**
 * Enum representing the state of a key slot (index) in
 * a key stream.
 */
typedef enum qkd_KSID_index_state {
    QKD_INDEX_EMPTY = 0,          // not filled w/ key material yet
    QKD_INDEX_READY = 1,          // key is ready to be used
    QKD_INDEX_SPENT = 2,          // key has been consumed already
    QKD_INDEX_OUT_OF_BOUNDS = 127 // index ouf of bounds
} qkd_KSID_index_state_t;

/**
 * Ringbuffer structure the QKMS mock-up uses to keep track
 * of "keys".  (The actual buffer is linear, of course.  Modulo
 * addressing is what makes it a ringbuffer.)
 *
 * The buffer holds ```size``` keys.  ```front_idx%size```
 * points to the first readable slot.  ```back_idx%size``` points
 * to the next writeable slot.  The buffer can be written to unless
 * it is full, i.e., unless ```key_count==size```.  The buffer can be
 * read from unless it is empty, i.e., unless ```key_count==0```.
 *
 * In fact, it is a bit of a strange ringbuffer in many ways.
 * - There is no write routine.  The ringbuffer tries to automagically
 *   fill itself during a read access.  See qkd_rb_read(.) for more
 *   details.
 * - The ringbuffer carries QoS related attributes such as chunk_size
 *   and TTF used in the process of automagically adjusting its contents
 *   during a read operation.  The same holds for a number of time tags
 *   (initialization time, last read access, last refill).
 */

typedef struct qkd_rb {
    time_t init_time;       // time of initialization
    time_t last_read_time;  // the time of the last successful read operation
    time_t last_fill_time;  // the time of the last refill
    unsigned size;          // number of keys (not bytes!) allocated
    unsigned chunk_size;    // number of bytes per key
    unsigned key_count;     // number of keys currently held
    unsigned spare_bytes;   // available bytes not yet stored in ringbuffer
    unsigned front_idx;     // first readable key slot
    unsigned back_idx;      // next writable key slot
    unsigned last_read_idx; // the idx of the last successful read operation
    unsigned key_gen_rate;  // key generation rate in byte/sec
    unsigned ttf;           // time to fetch (TTF) in seconds
    qkd_KSID_index_state_t *data; // the "ring"buffer
} qkd_rb_t;

/**
 * @brief Initialize a ringbuffer.
 * @param rb(inout) Non-NULL pointer to the ringbuffer structure to initialize.
 * @param size(in)  The number of keys the ringbuffer can hold, must be
 * non-zero.
 * @param key_gen_rate(in) Key generation rate in bytes per second.
 * @param chunk_size(in) Number of bytes per key.
 * @param ttf(in) TTF (time to fetch) in seconds.
 */
void qkd_rb_init(qkd_rb_t *rb, unsigned size, unsigned key_gen_rate,
                 unsigned chunk_size, unsigned ttf);

/**
 * @brief Frees resources allocated by this ringbuffer.
 * @param rb (inout) Non-NULL pointer to the ringbuffer
 */
void qkd_rb_dtor(qkd_rb_t *rb);

/**
 * @brief Read a key from a ringbuffer at the given index.
 * @param rb(inout) Non-NULL pointer to the ringbuffer
 * @param index(inout) Non-NULL (!) pointer to an index. Zero index value
 *        means "access oldest unconsumed key". The ringbuffer will assign
 *        (i.e., write back) the respective index value in this case.
 * @return Returns status at the given index. See qkd_KSID_index_state_t
 *         for details.
 *
 * This is what happens during a read operation.
 * 1) If space is available in the ringbuffer: check if we can fill the buffer.
 *     - First, add (NOW-last_fill_time)*key_gen_rate bytes to spare_bytes.
 *       Then, from there, fill the ringbuffer.  Adjust back_idx etc.
 * 2) TTF approximation: check if entries need to be deleted as their
 *    (approximated) TTF has expired.  If so: move front_idx etc.
 * 3) Again, try to fill free slots using spare_bytes.  (We do not attempt to
 *    re-run TTF approximation at this point.  This is good enough for a mock-up
 *    that is supposed to demonstrate that keys may get deleted if one waits too
 *    long after key creation.)
 * 4) Now, replace zero index with front_idx.
 * 5) If index==start_idx (this should be the normal case) AND
 *    buffer(index%size)==QKD_INDEX_READY then
 *     - Advance start_idx to the next unconsumed entry
 *     - Return QKD_INDEX_READY
 * 6) If index<start_idx OR index>=start_idx+size: return
 * QKD_INDEX_OUT_OF_BOUNDS 7) If data(index%size)==QKD_INDEX_READY: set entry to
 * QKD_INDEX_SPENT and return QKD_INDEX_READY. 8) Return data(index%size) (could
 * be QKD_INDEX_EMPTY or QKD_INDEX_SPENT)
 *
 * Here is how TTF approximation works:
 * - For proper TTF handling, we would need to implement a horizontal interface
 *   between A and B devices.  Lacking such an interface, we assume a "monotic
 *   peer" consuming a key every delta_T=runtime/front_idx seconds.
 * - Knowing that last_read_index was consumed at last_read_time, we can
 *   approximate the EOL time for key with other indices.  E.g., the key
 *   at position last_read_idx+x would expire at last_read_time+delta_T+TTF.
 */
qkd_KSID_index_state_t qkd_rb_read(qkd_rb_t *rb, unsigned *p_index);

#endif // ETSI_GS_QKD_004_RINGBUFFER_H
