#ifndef ETSI_GS_QKD_004_MODELS_H
#define ETSI_GS_QKD_004_MODELS_H

#include <stdint.h>
#include <uuid/uuid.h>

#ifdef __cplusplus__
extern "C" {
#endif

typedef struct qkd_bytes {
    uint32_t size;
    void *data;
} qkd_bytes_t;

void qkd_bytes_dtor(qkd_bytes_t *);

/**
 * @brief ETSI SG QKD 004 status codes.
 */
typedef enum qkd_status_code {
    qkd_status_code_successful = 0, /**< @brief Successful. */
    qkd_status_code_peer_not_connected =
        1, /**< @brief Successful connection, but peer not connected. */
    qkd_status_code_gk_insufficient_key_available =
        2, /**< @brief GET_KEY failed because insufficient key available. */
    qkd_status_code_gk_peer_not_connected =
        3, /**< @brief GET_KEY failed because peer application is not yet
              connected. */
    qkd_status_code_no_qkd_connection_available =
        4, /**< @brief No QKD connection available. */
    qkd_status_code_oc_ksid_already_in_use =
        5, /**< @brief OPEN_CONNECT failed because the "KSID" is already in use.
            */
    qkd_status_code_timeout_error = 6, /**< @brief TIMEOUT_ERROR The call failed
                                          because the specified TIMEOUT. */
    qkd_status_code_oc_qos_not_satisfiable =
        7, /**< @brief OPEN_CONNECT failed because requested QoS settings could
              not be met, counter proposal included in returnhas occurred. */
    qkd_status_code_gk_metadata_insufficient_size =
        8, /**< @brief GET_KEY failed because metadata field payload_size
              insufficient. Returned Metadata_size value holds minimum needed
              payload_size of metadata. */
    qkd_status_code_out_of_memory =
        9, /**< @brief Return or exit code in case memory allocation fails. */
    qkd_status_code_gk_index_of_out_bounds =
        10, /**< @brief GET_KEY failed due to index being out of bounds. */
    qkd_status_code_gk_key_spent =
        11, /**< @brief GET_KEY failed because key was already used and then
               removed from the server. */
    qkd_status_code_oc_no_route = 12, /**< @brief OPEN_CONNECT failed because no
                                         route to destination could be found*/
    qkd_status_code_oc_no_discovery_service =
        13, /**< @brief OPEN_CONNECT failed because no application discovery
               service was available. */
    qkd_status_code_oc_protocol_error =
        14, /**< @brief OPEN_CONNECT failed because of a protocol violation. */
    qkd_status_code_vendor_defined =
        0x80000000, /**< @brief Vendor-defined status start here. */
} qkd_status_code_t;

/**
 * @brief All status codes are 32 bits unsigned integers.
 */
typedef uint32_t qkd_status_t;

/**
 * @brief A qkd_uri is a sized string.
 */
typedef char *qkd_uri_t;

#define MIMETYPE_MAX 256
#define QKD_004_DEFAULT_PORTNUMBER 25575

/**
 * @brief A mimetype is a string with a maximum payload_size of 255 char + '\0'.
 */
typedef char *qkd_mimetype_t;

/**
 * @brief QoS parameters.
 */
typedef struct qkd_qos {
    uint32_t key_chunk_size;
    uint32_t max_bps;
    uint32_t min_bps;
    uint32_t capacity;
    uint32_t timeout;
    uint32_t ttf;
    qkd_mimetype_t oc_metadata_mimetype;
    qkd_mimetype_t gk_metadata_mimetype;
} qkd_qos_t;

void qkd_qos_dtor(qkd_qos_t *);

/**
 * @brief The qkd_metadata value buffer.
 */
typedef qkd_bytes_t qkd_metadata_t;

#define qkd_metadata_dtor qkd_bytes_dtor

/**
 * @brief   A key buffer.
 */
typedef qkd_bytes_t qkd_key_buffer_t;

#define qkd_key_buffer_dtor qkd_bytes_dtor

typedef struct qkd_open_connect_request {
    qkd_uri_t destination_kme_id;
    qkd_uri_t source_sae_id;
    qkd_uri_t destination_sae_id;
    qkd_qos_t qos;
    uuid_t key_stream_id;
    qkd_metadata_t metadata;
} qkd_open_connect_request_t;

void qkd_open_connect_request_dtor(qkd_open_connect_request_t *);

typedef struct qkd_open_connect_response {
    qkd_status_t status;
    qkd_qos_t qos;
    uuid_t key_stream_id;
    qkd_metadata_t metadata;
} qkd_open_connect_response_t;

void qkd_open_connect_response_dtor(qkd_open_connect_response_t *);

typedef struct qkd_get_key_request {
    uuid_t key_stream_id;
    uint32_t index;
    qkd_metadata_t metadata;
} qkd_get_key_request_t;

void qkd_get_key_request_dtor(qkd_get_key_request_t *);

typedef struct qkd_get_key_response {
    qkd_status_t status;
    uint32_t index;
    qkd_key_buffer_t key_buffer;
    qkd_metadata_t metadata;
} qkd_get_key_response_t;

void qkd_get_key_response_dtor(qkd_get_key_response_t *);

typedef struct qkd_close_request {
    uuid_t key_stream_id;
} qkd_close_request_t;

typedef struct qkd_close_response {
    qkd_status_t status;
} qkd_close_response_t;

typedef qkd_bytes_t qkd_echo_request_t;
#define qkd_echo_request_dtor qkd_bytes_dtor

typedef qkd_bytes_t qkd_echo_response_t;
#define qkd_echo_response_dtor qkd_bytes_dtor

// Service request/response specifier.
// Be careful when adding fields or changing the encoding;
// this needs to fit into a single byte.
enum qkd_service {
    QKD_SERVICE_ERROR = 0x00,
    QKD_SERVICE_OPEN_CONNECT_REQUEST = 0x02,
    QKD_SERVICE_OPEN_CONNECT_RESPONSE = 0x03,
    QKD_SERVICE_GET_KEY_REQUEST = 0x04,
    QKD_SERVICE_GET_KEY_RESPONSE = 0x05,
    QKD_SERVICE_CLOSE_REQUEST = 0x08,
    QKD_SERVICE_CLOSE_RESPONSE = 0x09,
    QKD_SERVICE_ECHO_REQUEST = 0xF0,  // for testing purposes
    QKD_SERVICE_ECHO_RESPONSE = 0xF1, // ditto
    QKD_SERVICE_CLOSE_SERVER = 0xFF,
};

typedef struct version {
    uint8_t major;
    uint8_t minor;
    uint8_t micro;
} version_t;

typedef struct qkd_header {
    uint32_t payload_size; // Comes first in order to avoid padding byte(s)
    version_t version;
    version_t protocol_version;
    uint8_t service;

} qkd_header_t;

#ifdef __cplusplus__
}
#endif

#endif // ETSI_GS_QKD_004_MODELS_H
