#include <assert.h>
#include <fcntl.h>
#include <limits.h>
#include <sys/stat.h>
#include <semaphore.h>
#include <signal.h>
#include <stdlib.h>

#include "qkd/cfg_parser.h"
#include "qkd/copy.h"
#include "qkd/interfaces.h"
#include "qkd/models.h"
#include "qkd/ssl_socket_server.h"
#include "qkd/sizes.h"

#include "examples/ringbuffer.h"

#include <openssl/evp.h>

typedef struct qkd_ksid_info {
    uuid_t key_stream_id;
    qkd_qos_t qos;
    qkd_uri_t source_sae_id;
    qkd_uri_t destination_sae_id;
    qkd_rb_t key_buffer;
    struct qkd_ksid_info *ksid_info_next;
} qkd_ksid_info_t;

static void qkd_ksid_info_dtor(qkd_ksid_info_t *k) {
    if (k == NULL) return;
    qkd_qos_dtor(&(k->qos));
    free(k->source_sae_id);
    k->source_sae_id = NULL;
    free(k->destination_sae_id);
    k->destination_sae_id = NULL;
    qkd_rb_dtor(&k->key_buffer);
}

typedef struct server_context_data_example {
    int key_rate;
    float qber;
    char *example_db_name;
    uuid_t node_uuid;
    qkd_ksid_info_t *ksid_info_linked_list;
} server_context_data_example_t;

// Name of the semaphore used to assure that a given
// serial-number + A/B combination is unique
static char *sem_name = NULL;
// The serial number
static long serial_number = 0;

#define CFG_NUM_KEYWORDS 8
/**
 * Table of (optional) configuration file entries.
 * - port_number = the server's port number
 * - max_num_KSIDs = maximum number of KSIDs
 * - max_num_sessions = maximum number of parallel sessions
 * - per_KSID_RB_size = per-KSID ringbuffer size in byte
 * - keygen_rate = key generation rate in byte per seconds (at most
 *   UINT_MAX/256)
 * - cert_pem = path to server certificate
 * - key_pem = path to server's private key
 * - peer_ca_pem = path to CA certificate, only set to enable client
 *   verification
 */
static cfg_kv_pair_t cfg_table[CFG_NUM_KEYWORDS + 1] = {
    {"port_number", NULL},
    {"max_num_KSIDs", NULL},
    {"max_num_sessions", NULL},
    {"per_KSID_RB_size", NULL},
    {"keygen_rate", NULL},
    {"cert_pem", NULL},
    {"key_pem", NULL},
    {"peer_ca_pem", NULL},
    {NULL, NULL} // the end-marker
};

// Configuration parameters and default values
const long default_keygen_rate = 128;
static uint16_t port_number = QKD_004_DEFAULT_PORTNUMBER;
static long max_num_KSIDs = 4;      // maximum number of KSIDs
static long max_num_sessions = 4;   // maximum number of parallel sessions
static long per_KSID_RB_size = 512; // per-KSID ringbuffer size in byte
static long keygen_rate =
    default_keygen_rate; // key generation rate in byte per seconds
// Certificate sending is MANDATORY for server side.
static const char *cert_pem = "../ssl/127.0.0.1.pem";
static const char *key_pem = "../ssl/127.0.0.1.key";
static const char *peer_ca_pem = NULL; // Set to, e.g., "../ssl/root_CA.pem" to
                                       // activate client verification.

// Remaining available key generation rate
static long available_keygen_rate = default_keygen_rate;

qkd_key_buffer_t gen_pseudorandom_key(uuid_t key_stream_id, qkd_uri_t source,
                                      qkd_uri_t destination, size_t key_size,
                                      uint32_t index) {
    qkd_key_buffer_t key = {key_size, malloc(key_size)};
    if (key.data == NULL) {
        fprintf(stderr, "ERROR: out of memory\n");
        exit(qkd_status_code_out_of_memory);
    }

    // We create a (reproducible) key from serial-number, KSID, and index
    // such that, if you modify and of these values you get (with very high
    // likelihood) a different key.
    const size_t msg_size = sizeof(uuid_t) + sizeof(long) + sizeof(uint32_t);
    char msg[msg_size];       // KSID | serial-number | index
    unsigned char digest[32]; // SHA256 hash of this "message"
    memcpy(msg, key_stream_id, sizeof(uuid_t));
    memcpy(msg + sizeof(uuid_t), &serial_number, sizeof(long));
    memcpy(msg + sizeof(uuid_t) + sizeof(long), &index, sizeof(uint32_t));
    EVP_Digest(msg, msg_size, digest, NULL, EVP_sha256(), NULL);
    for (size_t i = 0; i < key_size; i++)
        ((unsigned char *)key.data)[i] = digest[i % 32];
    return key;
}

void server_qkd_open_connect(qkd_uri_t destination_kme_id,
                             qkd_uri_t source_sae_id,
                             qkd_uri_t destination_sae_id, qkd_qos_t *qos,
                             uuid_t *key_stream_id, qkd_metadata_t *metadata,
                             qkd_status_t *status, void *app_context_data) {
    assert(qos != NULL);
    assert(key_stream_id != NULL);
    assert(status != NULL);
    assert(app_context_data != NULL);
    assert(metadata);

    // A "NULL" URI (mind that SAE/KME IDs are URIs) has its first byte set
    // to zero.  Passing a NULL pointer is not allowed.
    assert(destination_kme_id != NULL);
    assert(source_sae_id != NULL);
    assert(destination_kme_id != NULL);

    if (uuid_is_null(*key_stream_id))
        uuid_generate_random(
            *key_stream_id); // Generate random if KSID is
                             // 00000000-0000-0000-0000-000000000000

    // This mock-up ignores metadata
    if (metadata->size != 0) {
        printf("Ignoring metadata\n");
    }

    // This mock-up is not connected to a QKDN controller and cannot handle
    // non-NULL destination KME IDs.  We are essentially modeling a direct,
    // single-hop QKD link.
    if (*destination_kme_id != 0) {
        // Sorry, no way to determine a routing solution.
        *status = qkd_status_code_oc_no_route;
        return;
    } else {
        // This mock-up is not connected to an application discovery service
        // either and, hence ...
        if (*destination_sae_id != 0) {
            // ... we're bailing out.
            *status = qkd_status_code_oc_no_discovery_service;
            return;
        } else {
            if (*source_sae_id != 0) {
                // If the destination SAE ID is NULL then the source SAE ID
                // must be NULL too.
                *status = qkd_status_code_oc_protocol_error;
                return;
            }
        }
    }

    // Use or modify app_context_data
    server_context_data_example_t *config = app_context_data;
    qkd_ksid_info_t *ksid_info_list = config->ksid_info_linked_list;

    // Please note that this QKMS mock-up, lacking a horizontal connection to
    // its peer, cannot tell the difference between a first and a second
    // OPEN_CONNECT call referencing the same KSID.  =>  We are lacking some
    // checks here.  (If need be, we can add means to earmark certain KSIDs as
    // "already opened by my peer".)

    unsigned num_KSIDs = 0;
    while (ksid_info_list != NULL) { // Check KSID does not exist
        num_KSIDs++;
        if (uuid_compare(*key_stream_id, ksid_info_list->key_stream_id) == 0) {
            *status = qkd_status_code_oc_ksid_already_in_use;
            return;
        }
        ksid_info_list = ksid_info_list->ksid_info_next;
    }

    // Can we handle yet another key stream?
    if (num_KSIDs >= max_num_KSIDs) {
        // No, we can't -- we're maxed out already.
        // As this is an implementation-specific issue, we need to bail out
        // with a vendor-defined error code.
        *status = qkd_status_code_vendor_defined | 1;
        return;
    }

    // Error out if `Capacity * Key_chunk_size > per_KSID_RB_size` or
    // if we do not have enough unassigned key generation rate
    // available (we are looking for 1.5x min_bps -- our own policy,
    // not something the ETSI QKD 004 standard demands).
    if (qos->key_chunk_size * qos->capacity > per_KSID_RB_size ||
        available_keygen_rate / max_num_KSIDs == 0 ||
        3 * qos->min_bps / (2 * 8) > available_keygen_rate) {
        // Sigh.  The poor client must guess which QoS parameter caused this.
        *status = qkd_status_code_oc_qos_not_satisfiable;
        return;
    }

    // Now adjust the QoS parameters (QoS request -> QoS response)
    // to reflect what we guarantee.
    qos->capacity = per_KSID_RB_size / qos->key_chunk_size;
    qos->min_bps = (qos->min_bps != 0) ? 3 * qos->min_bps / (2 * 8)
                                       : available_keygen_rate / max_num_KSIDs;
    available_keygen_rate -= qos->min_bps;

    // Adding new ksid_info to shared data
    qkd_ksid_info_t *new_ksid_info = calloc(1, sizeof(qkd_ksid_info_t));
    if (new_ksid_info == NULL) {
        *status = qkd_status_code_out_of_memory;
        return;
    }
    uuid_copy(new_ksid_info->key_stream_id, *key_stream_id);
    qkd_copy_qos(&new_ksid_info->qos, qos);
    qkd_copy_uri(&new_ksid_info->source_sae_id, &source_sae_id);
    qkd_copy_uri(&new_ksid_info->destination_sae_id, &destination_sae_id);

    qkd_rb_init(&new_ksid_info->key_buffer,
                per_KSID_RB_size / qos->key_chunk_size, qos->min_bps,
                qos->key_chunk_size, qos->ttf);
    new_ksid_info->ksid_info_next = NULL;

    // Insert as the first element of the linked list
    if (config->ksid_info_linked_list == NULL) {
        config->ksid_info_linked_list = new_ksid_info;
    } else {
        new_ksid_info->ksid_info_next = config->ksid_info_linked_list;
        config->ksid_info_linked_list = new_ksid_info;
    }

    *status = qkd_status_code_successful;
}

void server_qkd_get_key(uuid_t key_stream_id, uint32_t *index,
                        qkd_key_buffer_t *key_buffer, qkd_metadata_t *metadata,
                        qkd_status_t *status, void *app_context_data) {
    assert(index != NULL);
    assert(key_buffer != NULL);
    assert(metadata != NULL);
    assert(status != NULL);
    assert(app_context_data != NULL);
    // Use or modify app_context_data
    server_context_data_example_t *config = app_context_data;

#define MAX_DEVICE_INFO_LEN 1024
    char device_info[MAX_DEVICE_INFO_LEN];
    snprintf(device_info, MAX_DEVICE_INFO_LEN,
             "{\"device_info\": [\"key_rate\": %d, \"qber\": %f]}",
             config->key_rate, config->qber);
    size_t data_size = qkd_sizeof_str(device_info);
    *metadata = (qkd_metadata_t){data_size, malloc(data_size)};
    if (metadata->data == NULL) {
        *status = qkd_status_code_out_of_memory;
        return;
    }
    memcpy((*metadata).data, device_info, data_size);

    qkd_ksid_info_t *ksid_info_list = config->ksid_info_linked_list;
    while (ksid_info_list != NULL) {
        if (uuid_compare(key_stream_id, ksid_info_list->key_stream_id) == 0)
            break;
        ksid_info_list = ksid_info_list->ksid_info_next;
    }

    if (ksid_info_list == NULL) { // KSID not found
        *status = qkd_status_code_gk_peer_not_connected;
        return;
    }

    // check the key buffer for availability of key material
    switch (qkd_rb_read(&ksid_info_list->key_buffer, index)) {
    case QKD_INDEX_EMPTY:
        *status = qkd_status_code_gk_insufficient_key_available;
        return;
    case QKD_INDEX_SPENT:
        *status = qkd_status_code_gk_key_spent;
        return;
    case QKD_INDEX_OUT_OF_BOUNDS:
        *status = qkd_status_code_gk_index_of_out_bounds;
        return;
    case QKD_INDEX_READY:
        break;
    }

    size_t key_size = ksid_info_list->qos.key_chunk_size;
    // THIS IS NOT A CRYPTOGRAPHICALLY SECURE KEY!
    qkd_key_buffer_t key_data = gen_pseudorandom_key(
        key_stream_id, ksid_info_list->source_sae_id,
        ksid_info_list->destination_sae_id, key_size, *index);
    *key_buffer = key_data;

    *status = qkd_status_code_successful;
}

void server_qkd_close(uuid_t key_stream_id, qkd_status_t *status,
                      void *app_context_data) {
    assert(status != NULL);
    assert(app_context_data != NULL);
    // Use or modify app_context_data
    server_context_data_example_t *config = app_context_data;
    qkd_ksid_info_t *ksid_info_list = config->ksid_info_linked_list,
                    *ksid_info_prev = NULL;
    while (ksid_info_list != NULL) {
        if (uuid_compare(key_stream_id, ksid_info_list->key_stream_id) == 0)
            break;
        ksid_info_prev = ksid_info_list;
        ksid_info_list = ksid_info_list->ksid_info_next;
    }

    if (ksid_info_list == NULL) { // KSID not found
        *status = qkd_status_code_no_qkd_connection_available;
        return;
    }

    // Reclaim the assigned key generation rate
    available_keygen_rate += ksid_info_list->key_buffer.key_gen_rate;

    if (ksid_info_prev == NULL)
        config->ksid_info_linked_list =
            ksid_info_list->ksid_info_next; // KSID found at the beginning of
                                            // the linked list
    else
        ksid_info_prev->ksid_info_next = ksid_info_list->ksid_info_next;

    qkd_ksid_info_dtor(ksid_info_list);
    free(ksid_info_list);

    *status = qkd_status_code_successful;
}

static void print_usage_and_exit(const char *program_name) {
    assert(program_name != NULL);
    fprintf(stderr, "Usage: %s <serial-number> A|B [<configuration-file>]\n",
            program_name);
    exit(1);
}

static void cleanup_at_exit() {
    if (sem_name != NULL) {
        sem_unlink(sem_name);
    }
}

static void signal_handler(const char *signal_name) {
    assert(signal_name != NULL);
#ifdef TRACE_SIGNAL_HANDLER
    // For printf-debugging purposes only.  (In general, using fprintf
    // inside a signal handler violates a number of good practices:
    // it's IO, potentially buffered, slow, and not signal-safe, to name
    // a just few).
    fprintf(stderr, "Received %s: exiting after cleanup\n", signal_name);
#endif
    cleanup_at_exit();
    exit(0);
}

static void sigint_handler() { signal_handler("SIGINT"); }
static void sigterm_handler() { signal_handler("SIGTERM"); }
static void sigkill_handler() { signal_handler("SIGKILL"); }
static void sigsegv_handler() { signal_handler("SIGSEGV"); }
static void sigbus_handler() { signal_handler("SIGBUS"); }

int main(int argc, char *argv[]) {
    // Check command-line arguments
    if (argc != 3 && argc != 4 ||
        strcmp(argv[2], "A") && strcmp(argv[2], "B")) {
        print_usage_and_exit(argv[0]);
    }
    char *invalid_chars = NULL;
    serial_number = strtol(argv[1], &invalid_chars, 10);

    if (argc == 4) {
        printf("Reading configuration file %s ...\n", argv[3]);
        if (cfg_from_file(cfg_table, argv[3]) != 0) {
            fprintf(stderr, "Loading the configuration file failed\n");
        } else {
            long val;

            int status = cfg_get_long(cfg_table, "port_number", &val);
            if (status == 0) {
                if (val < 0 || val > 65535) {
                    fprintf(stderr,
                            "Invalid port ID specified in configuration file "
                            "(valid range: 0..65535), skipping entry ...\n");
                } else {
                    port_number = val;
                    printf("Port number set to %d\n", port_number);
                }
            } else if (status == 2) {
                fprintf(stderr, "Skipping invalid port_number entry.\n");
            }

            status = cfg_get_long(cfg_table, "max_num_KSIDs", &val);
            if (status == 0 && val > 0) {
                max_num_KSIDs = val;
                printf("Maximum number of KSIDs set to %ld\n", max_num_KSIDs);
            } else if (status == 2) {
                fprintf(stderr, "Skipping invalid max_num_KSIDs entry.\n");
            }

            status = cfg_get_long(cfg_table, "max_num_sessions", &val);
            if (status == 0) {
                if (val == 0) {
                    fprintf(stderr, "Invalid max_num_sessions entry (must be "
                                    "larger than zero),"
                                    " skipping entry ...\n");
                } else {
                    max_num_sessions = val;
                    printf("Maximum number of TLS sessions set to %ld\n",
                           max_num_sessions);
                }
            } else if (status == 2) {
                fprintf(stderr, "Skipping invalid max_num_sessions entry.\n");
            }

            status = cfg_get_long(cfg_table, "per_KSID_RB_size", &val);
            if (status == 0) {
                per_KSID_RB_size = val;
                printf("Per-KSID ring buffer size set to %ld\n",
                       per_KSID_RB_size);
            } else if (status == 2) {
                fprintf(stderr, "Skipping invalid per_KSID_RB_size entry.\n");
            }

            status = cfg_get_long(cfg_table, "keygen_rate", &val);
            if (status == 0 && val > 0 && val <= UINT_MAX / 256) {
                keygen_rate = available_keygen_rate = val;
                printf("Key generation rate set to %ld\n", keygen_rate);
            } else if (status == 2) {
                fprintf(stderr, "Skipping invalid keygen_rate entry.\n");
            }

            const char *str_val = NULL;
            status = cfg_get_string(cfg_table, "cert_pem", &str_val);
            if (status == 0) {
                cert_pem = str_val;
                printf("Server certificate set to %s\n", cert_pem);
            } else {
                printf("Server certificate defaulting to %s\n", cert_pem);
            }

            status = cfg_get_string(cfg_table, "key_pem", &str_val);
            if (status == 0) {
                key_pem = str_val;
                printf("Server's private key set to %s\n", key_pem);
            } else {
                printf("Server's private key defaulting to %s\n", key_pem);
            }

            status = cfg_get_string(cfg_table, "peer_ca_pem", &str_val);
            if (status == 0) {
                peer_ca_pem = str_val;
                printf("Client verification enabled. Peer's CA certificate "
                       "set to %s\n",
                       peer_ca_pem);
            } else {
                printf("Client verification disabled.\n");
            }
        }
    }

    // Register cleanup routine and signal handlers
    atexit(cleanup_at_exit);
    signal(SIGINT, sigint_handler);
    signal(SIGTERM, sigterm_handler);
    signal(SIGKILL, sigkill_handler);
    signal(SIGSEGV, sigsegv_handler);
    signal(SIGBUS, sigbus_handler);

    // Make sure this is the one, and no one else has registered this
    // serial number + A/B combination.
    const char *prefix = "/ETSI_QKD_004_simulator_";
    sem_name = malloc(strlen(prefix) + strlen(argv[1]) + 2);
    if (sem_name == NULL) {
        fprintf(stderr, "ERROR: out of memory\n");
        exit(qkd_status_code_out_of_memory);
    }
    strcpy(sem_name, prefix);
    strcat(sem_name, argv[1]);
    strcat(sem_name, argv[2]);
    if (getenv("QKD_SERVER_FORCE_SEM_UNLINK")) {
        // Use this if, in spite of best efforts, the semaphore did not
        // get removed when the one and only exited.
        printf("Unlinking semaphore and exiting\n");
        sem_unlink(sem_name);
        return 0;
    }

    if (sem_open(sem_name, O_CREAT | O_EXCL, 0644, 0) == SEM_FAILED) {
        fprintf(stderr, "Error: the combination of serial number and A/B is "
                        "already in use\n");
        // Make sure the atexit handler does not unlink the semaphore used by
        // someone else
        free(sem_name);
        sem_name = NULL;
        exit(2);
    }

    // Register implementations of the interface functions
    qkd_open_connect = server_qkd_open_connect;
    qkd_get_key = server_qkd_get_key;
    qkd_close = server_qkd_close;

    // Optional (modify server_context_data_example_t to fit your preferences or
    // set it to NULL)
    server_context_data_example_t config = {
        32768, 0.49f, "database_file.db", {}, NULL};
    uuid_generate_random(config.node_uuid);

    // Supported TLS version range
    int min_proto_version = TLS1_3_VERSION;
    int max_proto_version = TLS1_3_VERSION;

    qkd_socket_server_init("127.0.0.1", port_number, cert_pem, key_pem,
                           peer_ca_pem, min_proto_version, max_proto_version,
                           max_num_sessions, &config);
    return 0;
}
