Server Example (QKMS mock-up)
=============================

This is a mock-up of a QKMS (quantum key management server).  You are
NOT looking at a full-fledged QKMS simulator; in particular not one
that correctly models the behavior of any QKD device or QKMS on the
market.

This mock-up models or, more correctly, _approximates_ the behavior of
a QKMS well enough to act as a stand-in in a couple of scenarios
including

- a QKD device pair yielding matching keys (given the applications on
  both sides use the same KSID and index (or, else, providing different
  key, which the applications should find problematic by the time it comes
  to using said keys),
- dealing with NULL and non-NULL KSIDs,
- dealing with NULL and non-NULL indices,
- modeling finite key rate (and, hence, signaling temporal unvailability
  of key material when the rate of consumption is too high),
- modeling finite storage capacity, i.e.
    * limiting the number of parallel KSIDs that can be handled and
    * limiting the number of unconsumed keys that can be stored per KSID
      which, for instance, can lead to "index out of bounds" errors if
      a client tries to access an index that is too far off from the
      smallest "unconsumed" index (note that this cannot happen if the
      client accesses the indices strictly sequentially, one after the
      other),
- keys being deleted due to TTF expiry,
- trying to access pre-consumed keys,
- simple QoS negotiation (key size, capacity, and min rate), and
- attempting to open a connection to an unknown destination.

We are trying to strike a balance between

- keeping the QKMS mock-up as simple as possible so that the
  focus remains on the ETSI QKD 004 interface and its behavior
  (in particular, we seek to avoid the implementation of complex
  key databases and horizontal QKMS-to-QKMS interfaces)
- and making it sufficiently powerful to work reasonably well in
  the scenarios described above.

The mock-up approximates the key management server of a
hypothical (not very well optimized) QKD device.  QKD devices
come in pairs.  This is modeled by a serial number and an
A/B identifier.  For a given serial number there is an 'A' and a
'B' device that are thought to be connected by a QKD link.

- Serial number and A/B identifier are given as command-line
  arguments, e.g.,
    `./server_example 1001 A`.
- Under the hood, named semaphores are used to ensure that a
  particular serial-number/A-B combination remains a singleton, i.e.,
  you can start
    `./server_example 1001 A`
  and
    `./server_example 1001 B`
  giving you a "connected" pair, but you cannot start
    `./server_example 1001 A`
  twice in parallel on the same machine.
- The keys generated by the QKMS mock-ups are not truly random but,
  instead, are the result of a key derivation function (KDF) that
  takes serial-number, KSID and index as parameters.  Hence:
    * Applications connected to an A/B pair of mock-up instances
      with the same serial-number will receive identical keys, if
      they use the same KSID and index on both sides.
    * Else, keys will be different (with a very high likelihood)
      reflecting what happens in the real world.
    * NOTE THAT THE MOCK-UPS DO NOT GENERATE CRYPTOGRAPHICALLY
      SECURE KEY MATERIAL.  They are good for integration testing
      purposes.  They are not suited for deployment.

There are a number of additional configuration parameters, all
of which have default values, that can optionally be specified
in a configuration file.

- `port_number`: The server's port number (default: 25575)
- `max_num_KSIDs`: The maximum number of KSIDs the server can
  handle in parallel (default: 4, quite low -- on purpose)
- `keygen_rate`: The key generation rate in bytes per second
  (default: 10, again quite small in order to provoke running
  into limitations)
- `per_KSID_RB_size`: The size (in bytes) of a ringbuffer-like storage
  that the server uses to store key material pertaining to a KSID
  (default: 512 byte, again quite small in order to be able to
  demonstrate index-out-of-bounds problems without much effort)
- `keygen_rate`: The key generation rate in bytes per second
  (default: 10, again quite small in order to provoke running
  into limitations)
- `cert_pem`: The path to the server's TLS certificate (default:
  "../ssl/127.0.0.1.pem")
- `key_pem`: The path to the server's private key (default:
  "../ssl/127.0.0.1.key")
- `peer_ca_pem`: The path to the CA certificate attesting the
  authenticity of the client's certificate (default: Unset, set it
  to, e.g., "../ssl/root_CA.pem" to enable client verification)

One can argue that a fixed a priori allocation of storage capacity
(`per_KSID_RB_size`) to each and every key stream is a bad idea, and
that on-demand allocation of key storage based on the `Capacity` QoS
parameter would result in a better QKMS architecture.  That is
certainly a fair comment.  However, the idea behind this mock-up is
not coming up with the optimal architecture but, instead, to (a)
keep it simple and (b) allow for easily triggering error cases such as
`Capacity * Key_chunk_size > per_KSID_RB_size`.

TTF _approximation_ (!) enables the QKMS mock-up to delete keys if
their TTF is expired.  For proper TTF handling, we would need a
full-fledged QKMS simulator implementing a horizontal interface
between A and B devices.  The mock-up does not do that.  Instead, it
assumes a _monotonic peer_ consuming a key at about every `delta_T =
runtime / number-of-keys-consumed seconds`.  Hence, it considers the
TTF of key with index `N` expired if it is being accessed later
than `t[N-1] + delta_T + TTF` (where `t[N-1]` is the access time of
the key with index `N-1`) or, if key `N-1` has not been consumed yet,
`t[N-M] + M * delta_T + TTF` if key `N-M` is the most recently
consumed key with an index smaller than `N`.

The QKMS mock-up uses the following algorithm for QoS negotation.

- Only Key_chunk_size, Min_bps, Capacity, and TTF are considered.
  Other QoS parameters are silently ignored.
- It errors out if
    `Key_chunk_size * Capacity > per_KSID_RB_size`.
  Else, it allocates the maximum capacity possible, i.e.,
    `Capacity = int(per_KSID_RB_size/Key_chunk_size)`
- It tries to honor the Min_bps requirement by allocating
  1.5x Min_Bps for the respective KSID if possible.  If not
  enough key generation bandwidth is available it will reject
  the request.
- In case of a NULL Min_bps requirement, it allocates
    `1/max_num_KSIDs * R`
  for the respective KSID where `R` is the remaining _unassigned_ key
  generation rate.  If no key generation bandwidth can be allocated at
  all for this KSID then it will reject the request.

Please note that this QKMS mock-up, lacking a horizontal connection to
its peer, cannot tell the difference between a first and a second
OPEN_CONNECT call referencing the same KSID.  => We are lacking some
checks here.  (If need be, we can add means to earmark certain KSIDs
are "already opened by my peer".)
