
#include <assert.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

#include "qkd/interfaces.h"
#include "qkd/ssl_socket.h"
#include "qkd/models.h"
#include "qkd/serializers.h"
#include "qkd/sizes.h"
#include "qkd/ssl_socket_client.h"
#include "qkd/copy.h"

void qkd_client_context_close_ssl(qkd_client_context_data_t *client_context) {
    if (client_context != NULL && client_context->ssl != NULL) {
        SSL_CTX *ctx = SSL_get_SSL_CTX(client_context->ssl);
        assert(ctx != NULL);
        SSL_free(client_context->ssl);
        SSL_CTX_free(ctx);
        client_context->ssl = NULL;
    }
}

/**
 * Perform a request.  Returns an empty data object if sending the request
 * failed. SSL object is created under the hood and initialized before first
 * use.
 */
static socket_data_t client_perform_request(socket_data_t request_data,
                                            void *client_context_data) {
    assert(client_context_data != NULL);
    qkd_client_context_data_t *cfg = client_context_data;

    if (cfg->ssl == NULL) {
        cfg->ssl = socket_client_init(
            cfg->host, cfg->port, cfg->cert_pem, cfg->key_pem, cfg->peer_ca_pem,
            cfg->min_proto_version, cfg->max_proto_version);
    }
#ifdef QKD_LOG_INFO
    fprintf(stdout,
            "[INFO] SSL connection on \"%s:%d\" using %s with cipher suite %s:",
            cfg->host, cfg->port, SSL_get_version(cfg->ssl),
            SSL_get_cipher(cfg->ssl));
#endif
    if (socket_ssl_send(cfg->ssl, request_data) != 0) {
        fprintf(stderr, "ERROR: sending request failed\n");
        socket_data_t no_data = {};
        return no_data; // Let the caller sort out the mess
    };
    socket_data_t header_data = socket_ssl_recv(cfg->ssl, sizeof(qkd_header_t));

    qkd_stream_t header_stream = {header_data.data,
                                  header_data.data + header_data.size};
    qkd_header_t header = qkd_header_from_stream(&header_stream);
    free(header_data.data);
#ifdef QKD_LOG_INFO
    switch (header.service) {
    case QKD_SERVICE_OPEN_CONNECT_RESPONSE:
        fprintf(stdout, " OPEN_CONNECT");
        break;
    case QKD_SERVICE_GET_KEY_RESPONSE:
        fprintf(stdout, " GET_KEY");
        break;
    case QKD_SERVICE_CLOSE_RESPONSE:
        fprintf(stdout, " CLOSE");
        break;
    case QKD_SERVICE_ECHO_RESPONSE:
        fprintf(stdout, " ECHO");
        break;
    }
    fprintf(stdout, "\n");
#endif
    socket_data_t retval = socket_ssl_recv(cfg->ssl, header.payload_size);
    return retval;
}

void client_qkd_open_connect(qkd_uri_t destination_kme_id,
                             qkd_uri_t source_sae_id,
                             qkd_uri_t destination_sae_id, qkd_qos_t *qos,
                             uuid_t *key_stream_id, qkd_metadata_t *metadata,
                             qkd_status_t *status, void *app_context_data) {
    assert(qos != NULL);
    assert(key_stream_id != NULL);
    assert(status != NULL);
    assert(app_context_data != NULL);
    assert(metadata != NULL);
    qkd_open_connect_request_t oc_request = {
        destination_kme_id, source_sae_id, destination_sae_id, *qos, {},
        *metadata};
    uuid_copy(oc_request.key_stream_id, *key_stream_id);

    qkd_header_t request_header = {qkd_sizeof_open_connect_request(oc_request),
                                   QKD_004_VERSION, QKD_004_PROTOCOL_VERSION,
                                   QKD_SERVICE_OPEN_CONNECT_REQUEST};

    size_t request_size = sizeof(qkd_header_t) + request_header.payload_size;
    void *request_buf = malloc(request_size);
    if (request_buf == NULL) {
        *status = qkd_status_code_out_of_memory;
        return;
    }
    qkd_stream_t request_stream = {request_buf, request_buf + request_size};
    qkd_header_to_stream(request_header, &request_stream);
    qkd_open_connect_request_to_stream(oc_request, &request_stream);

    // Send request
    socket_data_t response_data = client_perform_request(
        (socket_data_t){request_size, request_buf}, app_context_data);
    free(request_buf);
    request_buf = NULL;

    qkd_stream_t response_stream = {response_data.data,
                                    response_data.data + response_data.size};
    qkd_open_connect_response_t response =
        qkd_open_connect_response_from_stream(&response_stream);

    // Fill return values
    *status = response.status;
    uuid_copy(*key_stream_id, response.key_stream_id);
    qkd_copy_qos(qos, &response.qos);

    socket_data_dtor(&response_data);
    qkd_open_connect_response_dtor(&response);
}

void client_qkd_get_key(uuid_t key_stream_id, uint32_t *index,
                        qkd_key_buffer_t *key_buffer, qkd_metadata_t *metadata,
                        qkd_status_t *status, void *app_context_data) {
    assert(index != NULL);
    assert(key_buffer != NULL);
    assert(metadata != NULL);
    assert(status != NULL);
    assert(app_context_data != NULL);
    qkd_get_key_request_t gk_request = {{}, *index, *metadata};
    uuid_copy(gk_request.key_stream_id, key_stream_id);

    qkd_header_t request_header = {qkd_sizeof_get_key_request(gk_request),
                                   QKD_004_VERSION, QKD_004_PROTOCOL_VERSION,
                                   QKD_SERVICE_GET_KEY_REQUEST};

    size_t request_size = sizeof(qkd_header_t) + request_header.payload_size;
    void *request_buf = malloc(request_size);
    if (request_buf == NULL) {
        *status = qkd_status_code_out_of_memory;
        return;
    }
    qkd_stream_t request_stream = {request_buf, request_buf + request_size};

    qkd_header_to_stream(request_header, &request_stream);
    qkd_get_key_request_to_stream(gk_request, &request_stream);

    // Send request
    socket_data_t response_data = client_perform_request(
        (socket_data_t){request_size, request_buf}, app_context_data);
    free(request_buf);
    request_buf = NULL;

    qkd_stream_t response_stream = {response_data.data,
                                    response_data.data + response_data.size};
    qkd_get_key_response_t response =
        qkd_get_key_response_from_stream(&response_stream);

    // Fill return values
    // (We're moving the data out of the response and. hence, we do
    // not call the response's dtor.)
    *status = response.status;
    *index = response.index;
    *key_buffer = response.key_buffer;
    *metadata = response.metadata;

    socket_data_dtor(&response_data);
}

void client_qkd_close(uuid_t key_stream_id, qkd_status_t *status,
                      void *app_context_data) {
    assert(status != NULL);
    assert(app_context_data != NULL);
    qkd_close_request_t cl_request = {{}};
    uuid_copy(cl_request.key_stream_id, key_stream_id);

    qkd_header_t request_header = {qkd_sizeof_close_request(cl_request),
                                   QKD_004_VERSION, QKD_004_PROTOCOL_VERSION,
                                   QKD_SERVICE_CLOSE_REQUEST};

    size_t request_size = sizeof(qkd_header_t) + request_header.payload_size;
    void *request_buf = malloc(request_size);
    if (request_buf == NULL) {
        *status = qkd_status_code_out_of_memory;
        return;
    }
    qkd_stream_t request_stream = {request_buf, request_buf + request_size};
    qkd_header_to_stream(request_header, &request_stream);
    qkd_close_request_to_stream(cl_request, &request_stream);

    // Send request
    socket_data_t response_data = client_perform_request(
        (socket_data_t){request_size, request_buf}, app_context_data);
    free(request_buf);
    request_buf = NULL;

    qkd_stream_t response_stream = {response_data.data,
                                    response_data.data + response_data.size};
    qkd_close_response_t response =
        qkd_close_response_from_stream(&response_stream);

    // Fill return values
    *status = response.status;

    socket_data_dtor(&response_data);
}

void qkd_register_client_interface_routines() {
    // Register implementations of the interface functions
    qkd_open_connect = client_qkd_open_connect;
    qkd_get_key = client_qkd_get_key;
    qkd_close = client_qkd_close;
}

qkd_echo_response_t qkd_echo_response_struct(qkd_echo_request_t echo_request,
                                             void *app_context_data) {
    assert(app_context_data != NULL);

    qkd_header_t request_header = {qkd_sizeof_bytes(echo_request),
                                   QKD_004_VERSION, QKD_004_PROTOCOL_VERSION,
                                   QKD_SERVICE_ECHO_REQUEST};

    size_t request_size = sizeof(qkd_header_t) + request_header.payload_size;
    void *request_buf = malloc(request_size);
    if (request_buf == NULL) {
        fprintf(stderr, "ERROR: out of memory\n");
        exit(qkd_status_code_out_of_memory);
    }
    qkd_stream_t request_stream = {request_buf, request_buf + request_size};
    qkd_header_to_stream(request_header, &request_stream);
    qkd_bytes_data_to_stream(echo_request, &request_stream);

    // Send request
    socket_data_t response_data = client_perform_request(
        (socket_data_t){request_size, request_buf}, app_context_data);
    free(request_buf);
    request_buf = NULL;

    qkd_stream_t response_stream = {response_data.data,
                                    response_data.data + response_data.size};
    qkd_echo_response_t echo_response =
        qkd_bytes_data_from_stream(&response_stream);

    socket_data_dtor(&response_data);

#ifdef QKD_LOG_INFO
    fprintf(stdout, "[INFO] ");
    qkd_fprintf_bytes(stdout, echo_request, 0);
    fprintf(stdout, "[INFO] ");
    qkd_fprintf_bytes(stdout, echo_response, 0);
#endif

    return echo_response;
}
