#include <assert.h>
#include <string.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "qkd/ssl_socket.h"
#include "qkd/interfaces.h"
#include "qkd/serializers.h"
#include "qkd/sizes.h"
#include "qkd/ssl_socket_server.h"

#define EXIT_IF_OUT_OF_MEMORY(x)                                               \
    if (x == NULL) {                                                           \
        fprintf(stderr, "ERROR: out of memory\n");                             \
        exit(qkd_status_code_out_of_memory);                                   \
    }

/**
 * This servlet processes exactly one ETSI QKD004 command in a blocking fashion.
 */
static void qkd_servlet(SSL *client_ssl_socket, void *app_context_data,
                        server_status_t *server_status) {
    assert(client_ssl_socket != NULL);
    assert(app_context_data != NULL);
    assert(server_status != NULL);
    /* Prints connection INFO */
    int client_socket = SSL_get_fd(client_ssl_socket);
    struct sockaddr_in addr;
    socklen_t len = sizeof(addr);
    getpeername(client_socket, (struct sockaddr *)&addr, &len);

    // Header unpack
    socket_data_t header_data =
        socket_ssl_recv(client_ssl_socket, sizeof(qkd_header_t));
    qkd_stream_t header_stream =
        (qkd_stream_t){header_data.data, header_data.data + header_data.size};
    qkd_header_t header = qkd_header_from_stream(&header_stream);
    socket_data_dtor(&header_data);

#ifdef QKD_LOG_INFO
    fprintf(
        stdout,
        "[INFO] Client connection on \"%s:%d\" using %s with cipher suite %s: ",
        inet_ntoa(addr.sin_addr), ntohs(addr.sin_port),
        SSL_get_version(client_ssl_socket), SSL_get_cipher(client_ssl_socket));
    switch (header.service) {
    case QKD_SERVICE_OPEN_CONNECT_REQUEST:
        fprintf(stdout, "OPEN_CONNECT");
        break;
    case QKD_SERVICE_GET_KEY_REQUEST:
        fprintf(stdout, "GET_KEY");
        break;
    case QKD_SERVICE_CLOSE_REQUEST:
        fprintf(stdout, "CLOSE");
        break;
    case QKD_SERVICE_ECHO_REQUEST:
        fprintf(stdout, "ECHO");
    }
    fprintf(stdout, "\n");
#endif

    if (header.service == QKD_SERVICE_CLOSE_SERVER) { // Close server
        *server_status = SERVER_FINISHED;
        return;
    }

    // Payload unpack
    socket_data_t payload =
        socket_ssl_recv(client_ssl_socket, header.payload_size);
    qkd_stream_t payload_stream = {payload.data, payload.data + payload.size};

    qkd_header_t response_header = {0, QKD_004_VERSION,
                                    QKD_004_PROTOCOL_VERSION, 0};
    uint8_t *response_buf = NULL;
    qkd_stream_t response_stream = {};
    size_t response_size = sizeof(qkd_header_t);

    // Bail out in case of QKD version or protocol version mismatch.
    if (memcmp(&header.version, &QKD_004_VERSION, sizeof(version_t)) ||
        memcmp(&header.protocol_version, &QKD_004_PROTOCOL_VERSION,
               sizeof(version_t))) {
        // At this point, we don't even know for sure what type of service is
        // being requested.  Who knows?  The encryption of header.service may
        // have changed. The best we can do is dropping this request.
        return;
    }

    if (header.service == QKD_SERVICE_OPEN_CONNECT_REQUEST) {
        response_header.service = QKD_SERVICE_OPEN_CONNECT_RESPONSE;
        qkd_open_connect_request_t oc_request =
            qkd_open_connect_request_from_stream(&payload_stream);
        qkd_open_connect_response_t oc_response =
            qkd_open_connect_struct(oc_request, app_context_data);

        response_header.payload_size =
            qkd_sizeof_open_connect_response(oc_response);
        response_size += response_header.payload_size;
        response_buf = malloc(response_size);
        EXIT_IF_OUT_OF_MEMORY(response_buf);
        response_stream =
            (qkd_stream_t){response_buf, response_buf + response_size};

        qkd_header_to_stream(response_header, &response_stream);
        qkd_open_connect_response_to_stream(oc_response, &response_stream);

        qkd_open_connect_request_dtor(&oc_request);
        qkd_open_connect_response_dtor(&oc_response);
    } else if (header.service == QKD_SERVICE_GET_KEY_REQUEST) {
        response_header.service = QKD_SERVICE_GET_KEY_RESPONSE;
        qkd_get_key_request_t gk_request =
            qkd_get_key_request_from_stream(&payload_stream);
        qkd_get_key_response_t gk_response =
            qkd_get_key_struct(gk_request, app_context_data);

        response_header.payload_size = qkd_sizeof_get_key_response(gk_response);
        response_size += response_header.payload_size;
        response_buf = malloc(response_size);
        EXIT_IF_OUT_OF_MEMORY(response_buf);
        response_stream =
            (qkd_stream_t){response_buf, response_buf + response_size};

        qkd_header_to_stream(response_header, &response_stream);
        qkd_get_key_response_to_stream(gk_response, &response_stream);

        qkd_get_key_request_dtor(&gk_request);
        qkd_get_key_response_dtor(&gk_response);
    } else if (header.service == QKD_SERVICE_CLOSE_REQUEST) {
        response_header.service = QKD_SERVICE_CLOSE_RESPONSE;
        qkd_close_request_t cl_request =
            qkd_close_request_from_stream(&payload_stream);
        qkd_close_response_t cl_response =
            qkd_close_struct(cl_request, app_context_data);

        response_header.payload_size = qkd_sizeof_close_response(cl_response);
        response_size += response_header.payload_size;
        response_buf = malloc(response_size);
        EXIT_IF_OUT_OF_MEMORY(response_buf);
        response_stream =
            (qkd_stream_t){response_buf, response_buf + response_size};

        qkd_header_to_stream(response_header, &response_stream);
        qkd_close_response_to_stream(cl_response, &response_stream);
    } else if (header.service == QKD_SERVICE_ECHO_REQUEST) {
        response_header.service = QKD_SERVICE_ECHO_RESPONSE;
        qkd_echo_request_t echo_request =
            qkd_bytes_data_from_stream(&payload_stream);
        qkd_echo_response_t *echo_response = &echo_request;

        response_header.payload_size = qkd_sizeof_bytes(*echo_response);
        response_size += response_header.payload_size;
        response_buf = malloc(response_size);
        EXIT_IF_OUT_OF_MEMORY(response_buf);
        response_stream =
            (qkd_stream_t){response_buf, response_buf + response_size};

        qkd_header_to_stream(response_header, &response_stream);
        qkd_bytes_data_to_stream(*echo_response, &response_stream);
        qkd_echo_request_dtor(&echo_request);
    } else {
        fprintf(stderr, "Error: qkd service unknown.\n");
    }
    socket_data_dtor(&payload);

    // Send response
    if (socket_ssl_send(client_ssl_socket,
                        (socket_data_t){response_size, response_buf}) != 0) {
        fprintf(stderr, "ERROR: sending response failed\n");
    }
    free(response_buf);
}

void qkd_socket_server_init(const char *host, uint16_t port,
                            const char *cert_pem, const char *key_pem,
                            const char *ca_pem, int min_proto_version,
                            int max_proto_version, unsigned max_num_sessions,
                            void *server_context_data) {
    socket_server_init(host, port, cert_pem, key_pem, ca_pem, min_proto_version,
                       max_proto_version, max_num_sessions, qkd_servlet,
                       server_context_data);
}
