#include <assert.h>
#include <netinet/in.h>

#include "qkd/serializers.h"
#include "qkd/sizes.h"

#define EXIT_IF_OUT_OF_MEMORY(x)                                               \
    if (x == NULL) {                                                           \
        fprintf(stderr, "ERROR: out of memory\n");                             \
        exit(qkd_status_code_out_of_memory);                                   \
    }

void qkd_uint32_to_stream(uint32_t number, qkd_stream_t *stream) {
    assert(stream != NULL);
    assert(stream->pos < stream->end);
    number = htonl(number);

    size_t offset = sizeof(uint32_t);
    memcpy(stream->pos, &number, offset);
    stream->pos += offset;
    assert(stream->pos <= stream->end);
}

uint32_t qkd_uint32_from_stream(qkd_stream_t *stream) {
    assert(stream != NULL);
    assert(stream->pos < stream->end);
    uint32_t number;

    size_t offset = sizeof(uint32_t);
    memcpy(&number, stream->pos, offset);
    stream->pos += offset;
    assert(stream->pos <= stream->end);

    return ntohl(number);
}

void qkd_uuid_to_stream(uuid_t id, qkd_stream_t *stream) {
    assert(stream != NULL);
    assert(stream->pos < stream->end);
    size_t offset = sizeof(uuid_t);
    memcpy(stream->pos, id, offset);
    stream->pos += offset;
    assert(stream->pos <= stream->end);
}

void qkd_uuid_from_stream(uuid_tp id, qkd_stream_t *stream) {
    assert(stream != NULL);
    assert(stream->pos < stream->end);
    size_t offset = sizeof(uuid_t);
    uuid_copy(id, stream->pos);
    stream->pos += offset;
    assert(stream->pos <= stream->end);
}

void qkd_bytes_data_to_stream(qkd_bytes_t data, qkd_stream_t *stream) {
    assert(stream != NULL);
    assert(stream->pos < stream->end);
    size_t offset;
    qkd_uint32_to_stream(data.size, stream);
    offset = data.size;
    memcpy(stream->pos, data.data, offset);
    stream->pos += offset;
    assert(stream->pos <= stream->end);
}

qkd_bytes_t qkd_bytes_data_from_stream(qkd_stream_t *stream) {
    assert(stream != NULL);
    assert(stream->pos < stream->end);
    qkd_bytes_t data;
    size_t offset;
    data.size = qkd_uint32_from_stream(stream);
    offset = data.size;
    data.data = malloc(offset);
    EXIT_IF_OUT_OF_MEMORY(data.data);
    memcpy(data.data, stream->pos, offset);
    stream->pos += offset;
    assert(stream->pos <= stream->end);
    return data;
}

void qkd_str_to_stream(char *str, qkd_stream_t *stream) {
    assert(str != NULL);
    assert(stream != NULL);
    assert(stream->pos < stream->end);
    size_t offset;
    offset = qkd_sizeof_str(str);
    memcpy(stream->pos, str, offset);
    stream->pos += offset;
    assert(stream->pos <= stream->end);
}

char *qkd_str_from_stream(qkd_stream_t *stream) {
    assert(stream != NULL);
    assert(stream->pos < stream->end);
    char *str_ = stream->pos;
    size_t offset = qkd_sizeof_str(str_);
    char *str = malloc(offset);
    EXIT_IF_OUT_OF_MEMORY(str);
    memcpy(str, stream->pos, offset);
    stream->pos += offset;
    assert(stream->pos <= stream->end);
    return str;
}

void qkd_uri_to_stream(qkd_uri_t uri, qkd_stream_t *stream) {
    return qkd_str_to_stream(uri, stream);
}

qkd_uri_t qkd_uri_from_stream(qkd_stream_t *stream) {
    return (qkd_uri_t)qkd_str_from_stream(stream);
}

void qkd_mimetype_to_stream(char *mime, qkd_stream_t *stream) {
    return qkd_str_to_stream(mime, stream);
}

qkd_mimetype_t qkd_mimetype_from_stream(qkd_stream_t *stream) {
    return (qkd_mimetype_t)qkd_str_from_stream(stream);
}

void qkd_qos_to_stream(qkd_qos_t qos, qkd_stream_t *stream) {
    qkd_uint32_to_stream(qos.key_chunk_size, stream);
    qkd_uint32_to_stream(qos.max_bps, stream);
    qkd_uint32_to_stream(qos.min_bps, stream);
    qkd_uint32_to_stream(qos.capacity, stream);
    qkd_uint32_to_stream(qos.timeout, stream);
    qkd_uint32_to_stream(qos.ttf, stream);
    qkd_mimetype_to_stream(qos.oc_metadata_mimetype, stream);
    qkd_mimetype_to_stream(qos.gk_metadata_mimetype, stream);
}

qkd_qos_t qkd_qos_from_stream(qkd_stream_t *stream) {
    qkd_qos_t qos;
    qos.key_chunk_size = qkd_uint32_from_stream(stream);
    qos.max_bps = qkd_uint32_from_stream(stream);
    qos.min_bps = qkd_uint32_from_stream(stream);
    qos.capacity = qkd_uint32_from_stream(stream);
    qos.timeout = qkd_uint32_from_stream(stream);
    qos.ttf = qkd_uint32_from_stream(stream);
    qos.oc_metadata_mimetype = qkd_mimetype_from_stream(stream);
    qos.gk_metadata_mimetype = qkd_mimetype_from_stream(stream);

    return qos;
}

void qkd_key_buffer_to_stream(qkd_key_buffer_t data, qkd_stream_t *stream) {
    qkd_bytes_data_to_stream(data, stream);
}

qkd_key_buffer_t qkd_key_buffer_from_stream(qkd_stream_t *stream) {
    return (qkd_key_buffer_t)qkd_bytes_data_from_stream(stream);
}

void qkd_metadata_to_stream(qkd_key_buffer_t data, qkd_stream_t *stream) {
    qkd_bytes_data_to_stream(data, stream);
}

qkd_key_buffer_t qkd_metadata_from_stream(qkd_stream_t *stream) {
    return (qkd_key_buffer_t)qkd_bytes_data_from_stream(stream);
}

void qkd_open_connect_request_to_stream(
    qkd_open_connect_request_t open_connect_request, qkd_stream_t *stream) {
    qkd_uri_to_stream(open_connect_request.destination_kme_id, stream);
    qkd_uri_to_stream(open_connect_request.source_sae_id, stream);
    qkd_uri_to_stream(open_connect_request.destination_sae_id, stream);
    qkd_qos_to_stream(open_connect_request.qos, stream);
    qkd_uuid_to_stream(open_connect_request.key_stream_id, stream);
    qkd_metadata_to_stream(open_connect_request.metadata, stream);
}

qkd_open_connect_request_t
qkd_open_connect_request_from_stream(qkd_stream_t *stream) {
    qkd_open_connect_request_t open_connect_request = {};
    open_connect_request.destination_kme_id = qkd_uri_from_stream(stream);
    open_connect_request.source_sae_id = qkd_uri_from_stream(stream);
    open_connect_request.destination_sae_id = qkd_uri_from_stream(stream);
    open_connect_request.qos = qkd_qos_from_stream(stream);
    qkd_uuid_from_stream(open_connect_request.key_stream_id, stream);
    open_connect_request.metadata = qkd_metadata_from_stream(stream);

    return open_connect_request;
}

void qkd_open_connect_response_to_stream(
    qkd_open_connect_response_t open_connect_response, qkd_stream_t *stream) {
    qkd_uint32_to_stream(open_connect_response.status, stream);
    qkd_qos_to_stream(open_connect_response.qos, stream);
    qkd_uuid_to_stream(open_connect_response.key_stream_id, stream);
    qkd_metadata_to_stream(open_connect_response.metadata, stream);
}

qkd_open_connect_response_t
qkd_open_connect_response_from_stream(qkd_stream_t *stream) {
    qkd_open_connect_response_t open_connect_response = {};
    open_connect_response.status = qkd_uint32_from_stream(stream);
    open_connect_response.qos = qkd_qos_from_stream(stream);
    qkd_uuid_from_stream(open_connect_response.key_stream_id, stream);
    open_connect_response.metadata = qkd_metadata_from_stream(stream);

    return open_connect_response;
}

void qkd_get_key_request_to_stream(qkd_get_key_request_t get_key_request,
                                   qkd_stream_t *stream) {
    qkd_uuid_to_stream(get_key_request.key_stream_id, stream);
    qkd_uint32_to_stream(get_key_request.index, stream);
    qkd_metadata_to_stream(get_key_request.metadata, stream);
}

qkd_get_key_request_t qkd_get_key_request_from_stream(qkd_stream_t *stream) {
    qkd_get_key_request_t get_key_request = {};
    qkd_uuid_from_stream(get_key_request.key_stream_id, stream);
    get_key_request.index = qkd_uint32_from_stream(stream);
    get_key_request.metadata = qkd_metadata_from_stream(stream);

    return get_key_request;
}

void qkd_get_key_response_to_stream(qkd_get_key_response_t get_key_response,
                                    qkd_stream_t *stream) {
    qkd_uint32_to_stream(get_key_response.status, stream);
    qkd_uint32_to_stream(get_key_response.index, stream);
    qkd_key_buffer_to_stream(get_key_response.key_buffer, stream);
    qkd_metadata_to_stream(get_key_response.metadata, stream);
}

qkd_get_key_response_t qkd_get_key_response_from_stream(qkd_stream_t *stream) {
    qkd_get_key_response_t get_key_response;
    get_key_response.status = qkd_uint32_from_stream(stream);
    get_key_response.index = qkd_uint32_from_stream(stream);
    get_key_response.key_buffer = qkd_key_buffer_from_stream(stream);
    get_key_response.metadata = qkd_metadata_from_stream(stream);

    return get_key_response;
}

void qkd_close_request_to_stream(qkd_close_request_t close_request,
                                 qkd_stream_t *stream) {
    qkd_uuid_to_stream(close_request.key_stream_id, stream);
}

qkd_close_request_t qkd_close_request_from_stream(qkd_stream_t *stream) {
    qkd_close_request_t close_request = {};
    qkd_uuid_from_stream(close_request.key_stream_id, stream);

    return close_request;
}

void qkd_close_response_to_stream(qkd_close_response_t close_response,
                                  qkd_stream_t *stream) {
    qkd_uint32_to_stream(close_response.status, stream);
}

qkd_close_response_t qkd_close_response_from_stream(qkd_stream_t *stream) {
    qkd_close_response_t close_response;
    close_response.status = qkd_uint32_from_stream(stream);

    return close_response;
}

void qkd_header_to_stream(qkd_header_t header, qkd_stream_t *stream) {
    assert(stream != NULL);
    assert(stream->pos < stream->end);
    header.payload_size = htonl(header.payload_size);
    size_t offset = sizeof(qkd_header_t);
    memcpy(stream->pos, &header, offset);
    stream->pos += offset;
    assert(stream->pos <= stream->end);
}

qkd_header_t qkd_header_from_stream(qkd_stream_t *stream) {
    assert(stream != NULL);
    assert(stream->pos < stream->end);
    qkd_header_t header;
    size_t offset = sizeof(qkd_header_t);
    memcpy(&header, stream->pos, offset);
    stream->pos += offset;
    header.payload_size = ntohl(header.payload_size);
    assert(stream->pos <= stream->end);
    return header;
}
