#include <assert.h>
#include <openssl/err.h>
#include <arpa/inet.h>
#include <string.h>
#include <unistd.h>
#include <stdbool.h>
#include <poll.h>

#include "qkd/ssl_socket.h"

static void log_ssl() {
    unsigned long err;
    while ((err = ERR_get_error())) {
        char *err_msg = ERR_error_string(err, 0);
        if (!err_msg) return;
        ERR_print_errors_fp(stderr);
    }
}

static void socket_ssl_context_load_certs(SSL_CTX *ctx, const char *cert_file,
                                          const char *key_file) {
    assert(ctx != NULL);
    if (cert_file == NULL || key_file == NULL) return;

    if (SSL_CTX_use_certificate_file(ctx, cert_file, SSL_FILETYPE_PEM) <= 0) {
        ERR_print_errors_fp(stderr);
        SSL_CTX_free(ctx);
        exit(EXIT_FAILURE);
    }
    if (SSL_CTX_use_PrivateKey_file(ctx, key_file, SSL_FILETYPE_PEM) <= 0) {
        ERR_print_errors_fp(stderr);
        SSL_CTX_free(ctx);
        exit(EXIT_FAILURE);
    }
    if (!SSL_CTX_check_private_key(ctx)) // Verify private key
    {
        fprintf(stderr,
                "Private key does not match the certificate's private key\n");
        SSL_CTX_free(ctx);
        exit(EXIT_FAILURE);
    }
}

static void socket_ssl_context_load_peer_CA(SSL_CTX *ctx,
                                            const char *peer_ca_file,
                                            role_t role) {
    assert(ctx != NULL);
    if (peer_ca_file == NULL) {
        SSL_CTX_set_verify(ctx, SSL_VERIFY_NONE, NULL);
        return;
    }

    if (SSL_CTX_load_verify_locations(ctx, peer_ca_file, NULL) <= 0) {
        ERR_print_errors_fp(stderr);
        SSL_CTX_free(ctx);
        exit(EXIT_FAILURE);
    }
    SSL_CTX_set_verify(ctx,
                       (role == ROLE_CLIENT)
                           ? SSL_VERIFY_PEER
                           : SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT,
                       NULL);
}

static int do_not_allow_early_data(SSL *s, void *arg) {
    return 0; // Early data is not allowed
}

static SSL_CTX *ssl_context_create(role_t role, const char *cert_pem,
                                   const char *key_pem, const char *peer_ca_pem,
                                   int min_proto_version,
                                   int max_proto_version) {
    // As we are dealing with highest cybersecurity requirements, we allow only
    // cybersuites fulfilling the highest requirements.
    const char *allowed_ciphersuites = "TLS_AES_256_GCM_SHA384";

    if (role == ROLE_SERVER) {
        if (cert_pem == NULL || key_pem == NULL) {
            fprintf(stdout, "Error creating server context: certificate/key is "
                            "mandatory for server side.\n");
            exit(EXIT_FAILURE);
        }
    }

    switch (min_proto_version) {
    case TLS1_3_VERSION:
        break; // good choice
    default:
        fprintf(stdout,
                "Error creating OpenSSL context: at least TLS 1.3 required.\n");
        exit(EXIT_FAILURE);
    }
    switch (max_proto_version) {
    case TLS1_3_VERSION:
        break; // good choice
    default:
        fprintf(stdout,
                "Error creating OpenSSL context: at most TLS 1.3 supported.\n");
        exit(EXIT_FAILURE);
    }

    OpenSSL_add_ssl_algorithms();
    SSL_load_error_strings();
    const SSL_METHOD *method =
        role == ROLE_CLIENT ? TLS_client_method() : TLS_server_method();

    SSL_CTX *ctx = SSL_CTX_new(method);
    if (ctx == NULL) {
        fprintf(stdout, "Error creating context.\n");
        log_ssl();
        exit(EXIT_FAILURE);
    }

    if (role == ROLE_SERVER) {
        SSL_CTX_set_ecdh_auto(ctx, 1);
    }

    SSL_CTX_set_min_proto_version(ctx, min_proto_version);
    SSL_CTX_set_max_proto_version(ctx, max_proto_version);

    // We require at least TLS1_3_VERSION and, hence, call
    // SSL_CTX_set_ciphersuites(.) instead of SSL_set_cipher_list(.).
    if (SSL_CTX_set_ciphersuites(ctx, allowed_ciphersuites) <= 0) {
        ERR_print_errors_fp(stderr);
        SSL_CTX_free(ctx);
        exit(EXIT_FAILURE);
    }

    // For security reasons, we do not support session resumption.  Here, we are
    // taking steps to disable session resumption.
    // NB:
    // - This code only has the desired effect for TLS 1.3. (We do require
    // TLS 1.3.)
    // - Disabling session tickets altogether (setting num_tickets to zero in
    // the ctx)
    //   will have a negative impact on the server's ability to deal with
    //   post-handshake client certificates.  (Temporily setting num_tickets to
    //   1 prior to calling SSL_verify_client_post_handshake() should solve the
    //   issue.)
    //
    // Safety bolt #1: set number of session tickets to zero
    if (SSL_CTX_set_num_tickets(ctx, 0) == 0) {
        ERR_print_errors_fp(stderr);
        SSL_CTX_free(ctx);
        exit(EXIT_FAILURE);
    }
    // Safety bolt #2: Do not allow early data (= disable 0-RTT)
    SSL_CTX_set_allow_early_data_cb(ctx, do_not_allow_early_data, NULL);

    socket_ssl_context_load_certs(ctx, cert_pem, key_pem);
    socket_ssl_context_load_peer_CA(ctx, peer_ca_pem, role);

    return ctx;
}

static SSL *socket_ssl_init(SSL_CTX *ctx, int socket) {
    assert(ctx != NULL);
    SSL *ssl_socket = SSL_new(ctx);
    if (!ssl_socket) {
        fprintf(stdout, "Error creating SSL.\n");
        SSL_CTX_free(ctx);
        log_ssl();
        exit(EXIT_FAILURE);
    }

    SSL_set_fd(ssl_socket, socket);
    return ssl_socket;
}

// Determine whether we want to
// - Abort I/O (return value = SSL_ERROR_ZERO_RETURN)
// - Retry (all other return values)
// - Exit due to an unrecoverable error (happens inside this routine)
static int socket_ssl_check_errors(SSL *ssl, int len) {
    assert(ssl);
    if (len <= 0) {
        int err = SSL_get_error(ssl, len);
        switch (err) {
        case SSL_ERROR_WANT_WRITE:
        case SSL_ERROR_WANT_READ:
        case SSL_ERROR_ZERO_RETURN:
            return err;
        case SSL_ERROR_SYSCALL:
        case SSL_ERROR_SSL:
        default: // unrecoverable error
            ERR_print_errors_fp(stderr);
            exit(EXIT_FAILURE);
        }
    }
    return ~(SSL_ERROR_ZERO_RETURN); // Whatever, as long as it's not
                                     // SSL_ERROR_ZERO_RETURN
}

socket_data_t socket_ssl_recv(SSL *ssl, size_t num_bytes_to_read) {
    assert(ssl != NULL);
    assert(num_bytes_to_read != 0);
    size_t num_received = 0;
    socket_data_t msg = {.size = num_bytes_to_read};
    msg.data = malloc(num_bytes_to_read);
    if (msg.data == NULL) {
        fprintf(stderr, "ERROR: out of memory\n");
        exit(EXIT_FAILURE);
    }

    while (num_received < num_bytes_to_read) {
#ifdef QKD_LOG_DEBUG
        fprintf(stdout, "[DEBUG] Receiving");
#endif
        size_t len = SSL_read(ssl, msg.data + num_received,
                              (int)(num_bytes_to_read - num_received));
        if (len > 0) num_received += len;
#ifdef QKD_LOG_DEBUG
        fprintf(stdout, " (%zu/%zu)", num_received, num_bytes_to_read);
#endif
        if (socket_ssl_check_errors(ssl, (int)len) == SSL_ERROR_ZERO_RETURN) {
            // Oops.  Looks like the socket was closed by the peer.  With
            // TLS 1.3 this could indicate that the server needs the client to
            // send a valid certificate. All we can do right now is to adjust
            // the .size field of the return value and let the caller determine
            // which actions to take.
            msg.size = num_received;
#ifdef QKD_LOG_DEBUG
            fprintf(stdout, ": ");
            qkd_fprintf_key_buffer(stdout, msg);
            fprintf(stdout, "\n");
#endif
            return msg;
        }
    }

#ifdef QKD_LOG_DEBUG
    fprintf(stdout, ": ");
    qkd_fprintf_key_buffer(stdout, msg);
    fprintf(stdout, "\n");
#endif
    return msg;
}

int socket_ssl_send(SSL *ssl, socket_data_t data_bytes) {
    assert(ssl != NULL);
    int num_sent = 0;
#ifdef QKD_LOG_DEBUG
    fprintf(stdout, "[DEBUG] Sending (%d): ", data_bytes.size);
#endif
    if (data_bytes.size == 0) return 0;

    while (num_sent < data_bytes.size) {
        int len = SSL_write(ssl, data_bytes.data, (int)data_bytes.size);
        if (len > 0) num_sent += len;
        if (socket_ssl_check_errors(ssl, (int)len) == SSL_ERROR_ZERO_RETURN) {
            // Oops.  Looks like the socket was closed by the peer.  With
            // TLS 1.3 this could indicate that the server needs the client to
            // send a valid certificate.
#ifdef QKD_LOG_DEBUG
            fprintf(
                stdout,
                "SSL socket error: data_bytes sent payload_size (%u) differs"
                "from data_bytes sent (%d).\n",
                data_bytes.size, num_sent);
#endif
            return 1;
        }
    }
#ifdef QKD_LOG_DEBUG
    qkd_fprintf_key_buffer(stdout, data_bytes);
    fprintf(stdout, "\n");
#endif
    return 0;
}

static int socket_create_connection(const char *host, uint16_t port) {
    assert(host != NULL);
    int s;
    s = socket(AF_INET, SOCK_STREAM, 0);
    if (s < 0) {
        fprintf(stdout, "Error creating socket.\n");
        exit(EXIT_FAILURE);
    }
    struct sockaddr_in sa;
    memset(&sa, 0, sizeof(sa));
    sa.sin_family = AF_INET;
    sa.sin_addr.s_addr = inet_addr(host);
    sa.sin_port = htons(port);
    socklen_t socklen = sizeof(sa);
    if (connect(s, (struct sockaddr *)&sa, socklen)) {
        fprintf(stdout, "Error connecting to server \"%s:%d\"\n", host, port);
        exit(EXIT_FAILURE);
    }
    return s;
}

SSL *socket_client_init(const char *host, uint16_t port, const char *cert_pem,
                        const char *key_pem, const char *peer_ca_pem,
                        int min_proto_version, int max_proto_version) {
    int s = socket_create_connection(host, port);
    SSL_CTX *ctx =
        ssl_context_create(ROLE_CLIENT, cert_pem, key_pem, peer_ca_pem,
                           min_proto_version, max_proto_version);
    // For the sake of simplicity, we use this context only once,
    // namely for the SSL object we are about to create.
    // Hence, there is no need to keep track of the context object
    // (we can get it from the SSL object).
    SSL *ssl_socket = socket_ssl_init(ctx, s);
    int err = SSL_connect(ssl_socket);
    if (err <= 0) {
        fprintf(stdout, "Error creating SSL connection.  err=%x\n", err);
        log_ssl();
        fflush(stdout);
        exit(EXIT_FAILURE);
    }
    return ssl_socket;
}

int socket_create_server(const char *host, uint16_t port) {
    assert(host != NULL);
    int server_socket;
    struct sockaddr_in addr;
    server_socket = socket(PF_INET, SOCK_STREAM, 0);
    bzero(&addr, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = inet_addr(host);
    if (setsockopt(server_socket, SOL_SOCKET, SO_REUSEADDR, &(int){1},
                   sizeof(int)) < 0) {
        perror("Setting socket option SO_REUSEADDR failed");
        exit(EXIT_FAILURE);
    }
    if (bind(server_socket, (struct sockaddr *)&addr, sizeof(addr)) != 0) {
        perror("can't bind port");
        exit(EXIT_FAILURE);
    }
    if (listen(server_socket, 10) != 0) {
        perror("Can't configure listening port");
        exit(EXIT_FAILURE);
    }
    return server_socket;
}

void socket_server_init(const char *host, uint16_t port, const char *cert_pem,
                        const char *key_pem, const char *ca_pem,
                        int min_proto_version, int max_proto_version,
                        unsigned max_num_sessions,
                        void (*servlet)(SSL *ssl_socket,
                                        void *server_context_data,
                                        server_status_t *server_status),
                        void *server_context_data) {
    assert(host != NULL);
    assert(servlet != NULL);
    assert(max_num_sessions != 0);
    // Max number of file descriptors (one for the server socket)
    unsigned max_num_fds = max_num_sessions + 1;

    // Only root user have the permission to run the server socket
    if (port < 1024 && getuid() != 0) {
        fprintf(stdout,
                "This program need to be run as \"root\" for port %d"
                " (ports below 1024 are \"privileged ports\")",
                port);
        exit(EXIT_FAILURE);
    }
    fprintf(stdout, "[INFO] Running server on \"%s:%d\"...\n", host, port);
    SSL_CTX *ctx = ssl_context_create(ROLE_SERVER, cert_pem, key_pem, ca_pem,
                                      min_proto_version, max_proto_version);
    int server_socket = socket_create_server(host, port);
    server_status_t server_status = SERVER_LISTENING;

    struct pollfd pfds[max_num_fds]; // One extra, for the listener
    SSL *ssl_sockets[max_num_fds];
    for (int i = 0; i < max_num_fds; i++) pfds[i].events = POLLIN | POLLHUP;
    pfds[0].fd = server_socket;
    int num_fds = 1; // Number of file descriptors (sockets)

    while (server_status == SERVER_LISTENING) {
        int retval;
        do {
            retval = poll(pfds, num_fds, -1);
        } while (retval = -1 && errno == EINTR);
        if (retval == -1) {
            perror("Polling server sockets failed");
            exit(EXIT_FAILURE);
        }

        // Handle commands (in a blocking fashion), if any
        for (int i = 1; i < num_fds; i++) {
            if (pfds[i].revents) {
                pfds[i].revents = 0;
                char buffer[32];
                if (pfds[i].revents & POLLHUP ||
                    recv(pfds[i].fd, buffer, sizeof(buffer),
                         MSG_PEEK | MSG_DONTWAIT) == 0) {
                    SSL_shutdown(ssl_sockets[i]);
                    // In theory, SSL_shutdown() can return a zero indicating
                    // that SSL_shutdown() should be called for a bidirectional
                    // shutdown. This should not be necessary here as the peer
                    // slammed the door shut in the first place. In theory,
                    // there's also a -1 return value, which one receives in
                    // case of non-blocking sockets (which we don't use) or if
                    // there's a fatal error somewhere deep down that no one
                    // knows how to handle. Well, we pull the plug out of the
                    // sockets two lines down anyway. So, just calling
                    // SSL_shutdown() and moving on seems OK.
                    SSL_free(ssl_sockets[i]);
                    close(pfds[i].fd);
                    // Now we need to reshuffle and put the last entry into
                    // slot that just became available.  This may result in
                    // temporarily skipping a socket that has data to be read.
                    // Never mind, we'll catch it with the next poll().
                    ssl_sockets[i] = ssl_sockets[num_fds - 1];
                    pfds[i] = pfds[num_fds - 1];
                    num_fds--;
                } else {
                    servlet(ssl_sockets[i], server_context_data,
                            &server_status);
                }
            }
        }

        // Accept another connection?
        if (pfds[0].revents & POLLIN) {
            struct sockaddr_in addr;
            socklen_t len = sizeof(addr);
            int client_socket =
                accept(server_socket, (struct sockaddr *)&addr, &len);
            if (num_fds == max_num_fds) {
                // We cannot handle any more sessions
                // => close the socket right away
                close(client_socket);
            } else {
                SSL *ssl_socket = socket_ssl_init(ctx, client_socket);
                if (SSL_accept(ssl_socket) != 1) {
                    ERR_print_errors_fp(stderr);
                    close(client_socket);
                } else {
                    ssl_sockets[num_fds] = ssl_socket;
                    pfds[num_fds++].fd = client_socket;
                }
            }
        }
    }
    close(server_socket);
    SSL_CTX_free(ctx);
}
